#!/usr/bin/env python

import sys
import time
from collections import OrderedDict
from PyQt4 import QtGui, QtCore
import emcor_controller as ec
from pyrpci import RemotePCI

MULT_MICRO = 1000000.0
F_VOLTAGE = "{0:.3f}"
F_CURRENT = "{0:.3f}"
F_TEMP = "{0:.2f}"
F_CHANNEL = "{0:.6f}"

class MCORType(object):
    def __init__(self, fs_mon, fs_fdbck, fs_dac):
        """Input values should be in A, which are here transformed to uA."""
        self.fullscale_mon = fs_mon * MULT_MICRO
        self.fullscale_fdbck = fs_fdbck * MULT_MICRO
        self.fullscale_dac = fs_dac * MULT_MICRO

MCOR_TYPES = OrderedDict()
MCOR_TYPES["1"] = MCORType(2, 1, 1.02564)
MCOR_TYPES["2"] = MCORType(2, 2, 2.05128)
MCOR_TYPES["6"] = MCORType(12, 6, 6.15384)
MCOR_TYPES["12"] = MCORType(12, 12, 12.30768)
MCOR_TYPES["30"] = MCORType(30, 30, 30.7692)

class VoltageSpinBox(QtGui.QDoubleSpinBox):
    def __init__(self):
        super(VoltageSpinBox, self).__init__()
        self.setDecimals(6)
        self.setSingleStep(0.1)
        self.setMinimum(ec.ChannelsControl.DAC_V_MIN)
        self.setMaximum(ec.ChannelsControl.DAC_V_MAX)

class StyledLabel(QtGui.QLabel):
    def __init__(self):
        super(StyledLabel, self).__init__()
        self.setFrameShape(QtGui.QFrame.StyledPanel)

class ChannelsControlWidget(QtGui.QGroupBox):
    """Widget for showing channel readbacks and setting the values on all the channels."""
    def __init__(self):
        super(ChannelsControlWidget, self).__init__()
        self.setTitle("Channels control")
        self.setMinimumWidth(700)
        self._ch_sp_widgets = []
        self._ch_mon_widgets = []
        self._ch_fdbck_widgets = []
        self._ch_mon_r_widgets = []
        self._ch_fdbck_r_widgets = []
        
        lbl_all_ch = QtGui.QLabel("All channels (V):")
        lbl_all_ch_policy = lbl_all_ch.sizePolicy()
        lbl_all_ch_policy.setHorizontalPolicy(QtGui.QSizePolicy.Fixed)
        lbl_all_ch.setSizePolicy(lbl_all_ch_policy)
        self._spin_box = VoltageSpinBox()
        self._btn_set_all = QtGui.QPushButton("Set All")
        self._btn_set_all.setEnabled(False)
        self._btn_set_all.clicked.connect(self._change_sp_all)
        self._btn_set_min = QtGui.QPushButton("Set All to Min")
        self._btn_set_min.setEnabled(False)
        self._btn_set_min.clicked.connect(self._change_sp_min)
        self._btn_set_zero = QtGui.QPushButton("Set All to Zero")
        self._btn_set_zero.setEnabled(False)
        self._btn_set_zero.clicked.connect(self._change_sp_zero)
        self._btn_set_max = QtGui.QPushButton("Set All to Max")
        self._btn_set_max.setEnabled(False)
        self._btn_set_max.clicked.connect(self._change_sp_max)
        
        w_all_channels = QtGui.QWidget()
        layout_all = QtGui.QHBoxLayout()
        layout_all.addWidget(lbl_all_ch)
        layout_all.addWidget(self._spin_box)
        layout_all.addWidget(self._btn_set_all)
        layout_all.addWidget(self._btn_set_min)
        layout_all.addWidget(self._btn_set_zero)
        layout_all.addWidget(self._btn_set_max)
        w_all_channels.setLayout(layout_all)
        
        w_channels = QtGui.QWidget()
        layout_ch = QtGui.QGridLayout()
        layout_ch.setColumnStretch(1, 1)
        layout_ch.setColumnStretch(2, 1)
        layout_ch.setColumnStretch(3, 1)
        layout_ch.setColumnStretch(4, 1)
        layout_ch.setColumnStretch(5, 1)
        layout_ch.addWidget(QtGui.QLabel("Instantaneous (V)"), 0, 2, 1, 2, QtCore.Qt.AlignCenter)
        layout_ch.addWidget(QtGui.QLabel("Ripple (V)"), 0, 4, 1, 2, QtCore.Qt.AlignCenter)
        layout_ch.addWidget(QtGui.QLabel("Monitor"), 1, 2, QtCore.Qt.AlignCenter)
        layout_ch.addWidget(QtGui.QLabel("Feedback"), 1, 3, QtCore.Qt.AlignCenter)
        layout_ch.addWidget(QtGui.QLabel("Monitor"), 1, 4, QtCore.Qt.AlignCenter)
        layout_ch.addWidget(QtGui.QLabel("Feedback"), 1, 5, QtCore.Qt.AlignCenter)
        head = 2
        for i in range(ec.ChannelsControl.NUM_CHANNELS):
            lbl_ch = QtGui.QLabel("Channel " + str(i) + " (V):")
            layout_ch.addWidget(lbl_ch, i + head, 0, QtCore.Qt.AlignRight)
            
            w_ch = ChannelSPWidget(i)
            self._ch_sp_widgets.append(w_ch)
            layout_ch.addWidget(w_ch, i + head, 1)
            
            lbl_mon = StyledLabel()
            lbl_mon.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignCenter)
            self._ch_mon_widgets.append(lbl_mon)
            layout_ch.addWidget(lbl_mon, i + head, 2)
            
            lbl_fdbck = StyledLabel()
            lbl_fdbck.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignCenter)
            self._ch_fdbck_widgets.append(lbl_fdbck)
            layout_ch.addWidget(lbl_fdbck, i + head, 3)
            
            lbl_mon_ripple = StyledLabel()
            lbl_mon_ripple.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignCenter)
            self._ch_mon_r_widgets.append(lbl_mon_ripple)
            layout_ch.addWidget(lbl_mon_ripple, i + head, 4)
            
            lbl_fdbck_ripple = StyledLabel()
            lbl_fdbck_ripple.setAlignment(QtCore.Qt.AlignRight | QtCore.Qt.AlignCenter)
            self._ch_fdbck_r_widgets.append(lbl_fdbck_ripple)
            layout_ch.addWidget(lbl_fdbck_ripple, i + head, 5)
        w_channels.setLayout(layout_ch)
        
        size_policy = w_channels.sizePolicy()
        size_policy.setVerticalPolicy(QtGui.QSizePolicy.MinimumExpanding)
        w_channels.setSizePolicy(size_policy)
        
        layout = QtGui.QVBoxLayout()
        layout.addWidget(w_all_channels)
        layout.addWidget(w_channels)
        self.setLayout(layout)
    
    def _change_sp_all(self):
        for w_ch in self._ch_sp_widgets:
            w_ch.change_set_point(self._spin_box.value())
    
    def _change_sp_min(self):
        self._spin_box.setValue(self._spin_box.minimum())
        self._change_sp_all()
    
    def _change_sp_zero(self):
        self._spin_box.setValue(0)
        self._change_sp_all()
    
    def _change_sp_max(self):
        self._spin_box.setValue(self._spin_box.maximum())
        self._change_sp_all()
    
    def update(self, ch_control, mcor_type):
        for w_ch in self._ch_sp_widgets:
            w_ch.update(ch_control, mcor_type)
            
        if ch_control is not None:
            self._btn_set_all.setEnabled(True)
            self._btn_set_min.setEnabled(True)
            self._btn_set_zero.setEnabled(True)
            self._btn_set_max.setEnabled(True)
        
            for i in range(ec.ChannelsControl.NUM_CHANNELS):
                ua = ch_control.read_monitor(i)
                v = ec.remap_value(ua, -mcor_type.fullscale_mon, mcor_type.fullscale_mon, ec.ChannelsControl.ADC_V_MIN, ec.ChannelsControl.ADC_V_MAX)
                self._ch_mon_widgets[i].setText(F_CHANNEL.format(v))
                
                ua = ch_control.read_feedback(i)
                v = ec.remap_value(ua, -mcor_type.fullscale_fdbck, mcor_type.fullscale_fdbck, ec.ChannelsControl.ADC_V_MIN, ec.ChannelsControl.ADC_V_MAX)
                self._ch_fdbck_widgets[i].setText(F_CHANNEL.format(v))
                
                ua = ch_control.read_monitor_ripple(i)
                v = ec.remap_value(ua, -mcor_type.fullscale_mon, mcor_type.fullscale_mon, ec.ChannelsControl.ADC_V_MIN, ec.ChannelsControl.ADC_V_MAX)
                self._ch_mon_r_widgets[i].setText(F_CHANNEL.format(v))
                
                ua = ch_control.read_feedback_ripple(i)
                v = ec.remap_value(ua, -mcor_type.fullscale_fdbck, mcor_type.fullscale_fdbck, ec.ChannelsControl.ADC_V_MIN, ec.ChannelsControl.ADC_V_MAX)
                self._ch_fdbck_r_widgets[i].setText(F_CHANNEL.format(v))
        else:
            self._btn_set_all.setEnabled(False)
            self._btn_set_min.setEnabled(False)
            self._btn_set_zero.setEnabled(False)
            self._btn_set_max.setEnabled(False)
            
            for i in range(ec.ChannelsControl.NUM_CHANNELS):
                self._ch_mon_widgets[i].setText("")
                self._ch_fdbck_widgets[i].setText("")
                self._ch_mon_r_widgets[i].setText("")
                self._ch_fdbck_r_widgets[i].setText("")
    
class ChannelSPWidget(QtGui.QWidget):
    """Widget for writing setpoints to just one channel."""
    def __init__(self, num):
        super(ChannelSPWidget, self).__init__()
        self._channel_num = num
        self._ch_control = None
        self._fs_dac = 0
        
        self._spin_box = VoltageSpinBox()
        
        self._btn_set = QtGui.QPushButton("Set")
        self._btn_set.setEnabled(False)
        self._btn_set.clicked.connect(self._change_set_point)
        
        layout = QtGui.QHBoxLayout()
        layout.addWidget(self._spin_box)
        layout.addWidget(self._btn_set)
        layout.setMargin(0)
        self.setLayout(layout)
    
    def change_set_point(self, value):
        self._spin_box.setValue(value)
        self._change_set_point()
        
    def _change_set_point(self):
        """Can only write the setpoint if the widget was updated before."""
        if self._ch_control is not None:
            v = self._spin_box.value()
            ua = ec.remap_value(v, ec.ChannelsControl.DAC_V_MIN, ec.ChannelsControl.DAC_V_MAX, -self._fs_dac, self._fs_dac)
            self._ch_control.write_set_point(self._channel_num, ua)
    
    def update(self, ch_control, mcor_type):
        """Saves the ch_control for later usage when setting the setpoints."""
        if ch_control:
            if self._ch_control is None:
                self._fs_dac = int(mcor_type.fullscale_dac / MULT_MICRO) * MULT_MICRO
                
                # XXX: configure bit currently bugged, just always configure all
                #if not ch_control.read_configure_bit(self._channel_num):
                ch_control.write_fullscale_adc_monitor(self._channel_num, mcor_type.fullscale_mon)
                ch_control.write_fullscale_adc_feedback(self._channel_num, mcor_type.fullscale_fdbck)
                ch_control.write_fullscale_dac(self._channel_num, mcor_type.fullscale_dac)
                ch_control.write_ramp_mode(self._channel_num, True)
                ch_control.write_configure_bit(self._channel_num, True)
                
                self._btn_set.setEnabled(True)
        else:
            self._btn_set.setEnabled(False)
        self._ch_control = ch_control

class SysInfoWidget(QtGui.QGroupBox):
    def __init__(self):
        super(SysInfoWidget, self).__init__()
        self.setTitle("System Info")
        
        self._lbl_sysid = StyledLabel()
        self._lbl_subtype = StyledLabel()
        self._lbl_frmver = StyledLabel()
        self._lbl_frmdate = StyledLabel()
        
        layout = QtGui.QFormLayout()
        layout.setLabelAlignment(QtCore.Qt.AlignRight)
        layout.addRow("System ID:", self._lbl_sysid)
        layout.addRow("Sub Type:", self._lbl_subtype)
        layout.addRow("Firmware version:", self._lbl_frmver)
        layout.addRow("Firmware date:", self._lbl_frmdate)
        self.setLayout(layout)
    
    def update(self, sys_info):
        if sys_info is not None:
            self._lbl_sysid.setText(sys_info.read_system_id())
            self._lbl_subtype.setText(sys_info.read_sub_type())
            self._lbl_frmver.setText(sys_info.read_firmware_version())
            self._lbl_frmdate.setText(sys_info.read_firmware_date())
        else:
            self._lbl_sysid.setText("")
            self._lbl_subtype.setText("")
            self._lbl_frmver.setText("")
            self._lbl_frmdate.setText("")

class XilinxSysMonitorWidget(QtGui.QGroupBox):
    def __init__(self):
        super(XilinxSysMonitorWidget, self).__init__()
        self.setTitle("Xilinx System Monitor")
        
        self._lbl_temp_cur = StyledLabel()
        self._lbl_temp_min = StyledLabel()
        self._lbl_temp_max = StyledLabel()
        
        self._lbl_vccint_cur = StyledLabel()
        self._lbl_vccint_min = StyledLabel()
        self._lbl_vccint_max = StyledLabel()
        
        self._lbl_vccaux_cur = StyledLabel()
        self._lbl_vccaux_min = StyledLabel()
        self._lbl_vccaux_max = StyledLabel()
        
        layout = QtGui.QGridLayout()
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(2, 1)
        layout.setColumnStretch(3, 1)
        layout.addWidget(QtGui.QLabel("Current"), 0, 1, QtCore.Qt.AlignCenter)
        layout.addWidget(QtGui.QLabel("Min"), 0, 2, QtCore.Qt.AlignCenter)
        layout.addWidget(QtGui.QLabel("Max"), 0, 3, QtCore.Qt.AlignCenter)
        layout.addWidget(QtGui.QLabel("Xilinx Temp (C):"), 1, 0, QtCore.Qt.AlignRight)
        layout.addWidget(self._lbl_temp_cur, 1, 1)
        layout.addWidget(self._lbl_temp_min, 1, 2)
        layout.addWidget(self._lbl_temp_max, 1, 3)
        layout.addWidget(QtGui.QLabel("Vcc Int (V):"), 2, 0, QtCore.Qt.AlignRight)
        layout.addWidget(self._lbl_vccint_cur, 2, 1)
        layout.addWidget(self._lbl_vccint_min, 2, 2)
        layout.addWidget(self._lbl_vccint_max, 2, 3)
        layout.addWidget(QtGui.QLabel("Vcc Aux (V):"), 3, 0, QtCore.Qt.AlignRight)
        layout.addWidget(self._lbl_vccaux_cur, 3, 1)
        layout.addWidget(self._lbl_vccaux_min, 3, 2)
        layout.addWidget(self._lbl_vccaux_max, 3, 3)
        self.setLayout(layout)
    
    def update(self, sys_mon):
        if sys_mon is not None:
            self._lbl_temp_cur.setText(F_TEMP.format(sys_mon.read_temp_cur()))
            self._lbl_temp_min.setText(F_TEMP.format(sys_mon.read_temp_min()))
            self._lbl_temp_max.setText(F_TEMP.format(sys_mon.read_temp_max()))
            self._lbl_vccint_cur.setText(F_VOLTAGE.format(sys_mon.read_vint_cur()))
            self._lbl_vccint_min.setText(F_VOLTAGE.format(sys_mon.read_vint_min()))
            self._lbl_vccint_max.setText(F_VOLTAGE.format(sys_mon.read_vint_max()))
            self._lbl_vccaux_cur.setText(F_VOLTAGE.format(sys_mon.read_vaux_cur()))
            self._lbl_vccaux_min.setText(F_VOLTAGE.format(sys_mon.read_vaux_min()))
            self._lbl_vccaux_max.setText(F_VOLTAGE.format(sys_mon.read_vaux_max()))
        else:
            self._lbl_temp_cur.setText("")
            self._lbl_temp_min.setText("")
            self._lbl_temp_max.setText("")
            self._lbl_vccint_cur.setText("")
            self._lbl_vccint_min.setText("")
            self._lbl_vccint_max.setText("")
            self._lbl_vccaux_cur.setText("")
            self._lbl_vccaux_min.setText("")
            self._lbl_vccaux_max.setText("")
        
class VoltageMonitorWidget(QtGui.QGroupBox):
    def __init__(self):
        super(VoltageMonitorWidget, self).__init__()
        self.setTitle("Voltage Monitor")
        
        self._lbl_v15 = StyledLabel()
        self._lbl_v12 = StyledLabel()
        self._lbl_v5 = StyledLabel()
        self._lbl_v33 = StyledLabel()
        self._lbl_v33io = StyledLabel()
        self._lbl_v25 = StyledLabel()
        self._lbl_v1 = StyledLabel()
        self._lbl_vm15 = StyledLabel()
        
        self._lbl_c15 = StyledLabel()
        self._lbl_c12 = StyledLabel()
        self._lbl_c5 = StyledLabel()
        self._lbl_c33 = StyledLabel()
        self._lbl_c33io = StyledLabel()
        self._lbl_c25 = StyledLabel()
        self._lbl_c1 = StyledLabel()
        self._lbl_cm15 = StyledLabel()
        
        volts_widget = QtGui.QWidget()
        volts_layout = QtGui.QGridLayout()
        volts_layout.setColumnStretch(1, 1)
        volts_layout.setColumnStretch(2, 1)
        volts_layout.addWidget(QtGui.QLabel("Voltage (V)"), 0, 1, QtCore.Qt.AlignCenter)
        volts_layout.addWidget(QtGui.QLabel("Current (A)"), 0, 2, QtCore.Qt.AlignCenter)
        volts_layout.addWidget(QtGui.QLabel("+15.0V:"), 1, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v15, 1, 1)
        volts_layout.addWidget(self._lbl_c15, 1, 2)
        volts_layout.addWidget(QtGui.QLabel("+12.0V:"), 2, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v12, 2, 1)
        volts_layout.addWidget(self._lbl_c12, 2, 2)
        volts_layout.addWidget(QtGui.QLabel("+5.0V:"), 3, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v5, 3, 1)
        volts_layout.addWidget(self._lbl_c5, 3, 2)
        volts_layout.addWidget(QtGui.QLabel("+3.3V:"), 4, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v33, 4, 1)
        volts_layout.addWidget(self._lbl_c33, 4, 2)
        volts_layout.addWidget(QtGui.QLabel("+3.3V IO:"), 5, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v33io, 5, 1)
        volts_layout.addWidget(self._lbl_c33io, 5, 2)
        volts_layout.addWidget(QtGui.QLabel("+2.5V:"), 6, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v25, 6, 1)
        volts_layout.addWidget(self._lbl_c25, 6, 2)
        volts_layout.addWidget(QtGui.QLabel("+1.0V:"), 7, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_v1, 7, 1)
        volts_layout.addWidget(self._lbl_c1, 7, 2)
        volts_layout.addWidget(QtGui.QLabel("-15.0V:"), 8, 0, QtCore.Qt.AlignRight)
        volts_layout.addWidget(self._lbl_vm15, 8, 1)
        volts_layout.addWidget(self._lbl_cm15, 8, 2)
        volts_widget.setLayout(volts_layout)
        
        size_policy = volts_widget.sizePolicy()
        size_policy.setVerticalPolicy(QtGui.QSizePolicy.MinimumExpanding)
        volts_widget.setSizePolicy(size_policy)
        
        self._lbl_temp = StyledLabel()
        temp_widget = QtGui.QWidget()
        temp_layout = QtGui.QFormLayout()
        temp_layout.setLabelAlignment(QtCore.Qt.AlignRight)
        temp_layout.addRow("Temperature (C):", self._lbl_temp)
        temp_widget.setLayout(temp_layout)
        
        layout = QtGui.QVBoxLayout()
        layout.addWidget(volts_widget)
        layout.addWidget(temp_widget)
        layout.setMargin(0)
        self.setLayout(layout)
    
    def update(self, v_mon, sys_mon):
        if v_mon is not None and sys_mon is not None:
            self._lbl_v15.setText(F_VOLTAGE.format(v_mon.read_voltage_15()))
            self._lbl_v12.setText(F_VOLTAGE.format(v_mon.read_voltage_12()))
            self._lbl_v5.setText(F_VOLTAGE.format(v_mon.read_voltage_5()))
            self._lbl_v33.setText(F_VOLTAGE.format(v_mon.read_voltage_33()))
            self._lbl_v33io.setText(F_VOLTAGE.format(v_mon.read_voltage_33io()))
            self._lbl_v25.setText(F_VOLTAGE.format(sys_mon.read_vaux_cur()))
            self._lbl_v1.setText(F_VOLTAGE.format(sys_mon.read_vint_cur()))
            self._lbl_vm15.setText(F_VOLTAGE.format(v_mon.read_voltage_m15()))
            
            self._lbl_c15.setText(F_CURRENT.format(v_mon.read_current_15()))
            self._lbl_c12.setText(F_CURRENT.format(v_mon.read_current_12()))
            self._lbl_c5.setText(F_CURRENT.format(v_mon.read_current_5()))
            self._lbl_c33.setText(F_CURRENT.format(v_mon.read_current_33()))
            self._lbl_c33io.setText(F_CURRENT.format(v_mon.read_current_33io()))
            self._lbl_c25.setText(F_CURRENT.format(v_mon.read_current_25()))
            self._lbl_c1.setText(F_CURRENT.format(v_mon.read_current_1()))
            self._lbl_cm15.setText(F_CURRENT.format(v_mon.read_current_m15()))
            
            self._lbl_temp.setText(F_TEMP.format(v_mon.read_temp()))
        else:
            self._lbl_v15.setText("")
            self._lbl_v12.setText("")
            self._lbl_v5.setText("")
            self._lbl_v33.setText("")
            self._lbl_v33io.setText("")
            self._lbl_v25.setText("")
            self._lbl_v1.setText("")
            self._lbl_vm15.setText("")
            
            self._lbl_c15.setText("")
            self._lbl_c12.setText("")
            self._lbl_c5.setText("")
            self._lbl_c33.setText("")
            self._lbl_c33io.setText("")
            self._lbl_c25.setText("")
            self._lbl_c1.setText("")
            self._lbl_cm15.setText("")
            
            self._lbl_temp.setText("")

class UpdateThreadObj(QtCore.QObject):
    """Background thread for updating Xilinx system monitor,
    voltage monitor and channels display. It runs continuously
    when the connection to the controller is established.
    """
    finished = QtCore.pyqtSignal()
    
    SLEEP_SEC = 0.05
    MIN_UPDATE_TIME_SEC = 1
    
    def __init__(self, rpci, mcor_type, w_sysinfo, w_sysmon, w_vmon, w_channels):
        QtCore.QObject.__init__(self)
        self._stop = False
        
        self._sysinfo = ec.SystemInfo(rpci)
        self._sysmon = ec.XilinxSystemMonitor(rpci)
        self._vmon = ec.VoltageMonitor(rpci)
        self._cc = ec.ChannelsControl(rpci)
        
        self._mcor_type = mcor_type
        
        self._w_sys_info = w_sysinfo
        self._w_xilinx_sys_monitor = w_sysmon
        self._w_v_monitor = w_vmon
        self._w_channels = w_channels

        self._exception = None
        self.finished.connect(self._on_finish)

    def do_run(self):
        try:
            self._w_sys_info.update(self._sysinfo)
            
            while not self._stop:
                start_time = time.time()
                self._w_xilinx_sys_monitor.update(self._sysmon)
                if self._stop:
                    break
                self._w_v_monitor.update(self._vmon, self._sysmon)
                if self._stop:
                    break
                self._w_channels.update(self._cc, self._mcor_type)
                if self._stop:
                    break
                
                duration_time = time.time() - start_time
                while duration_time <= UpdateThreadObj.MIN_UPDATE_TIME_SEC:
                    time.sleep(UpdateThreadObj.SLEEP_SEC)
                    if self._stop:
                        break
                    duration_time = time.time() - start_time
        except Exception as e:
            self._exception = e
        finally:
            self.finished.emit()
    
    def do_stop(self):
        self._stop = True
    
    def _on_finish(self):
        if self._exception:
            msg_box = QtGui.QMessageBox()
            msg_box.setText("Communication error!")
            msg_box.setInformativeText(str(self._exception))
            msg_box.setIcon(QtGui.QMessageBox.Critical)
            msg_box.exec_()
        
class MainWindow(QtGui.QMainWindow):
    def __init__(self):
        super(MainWindow, self).__init__()

        self._rpci = None
        self._exec_thread = None
        self._exec_obj = None
        self._exec_mutex = QtCore.QMutex()
        main_widget = QtGui.QWidget()
        
        self._combo_type = QtGui.QComboBox()
        for t in MCOR_TYPES.keys():
            self._combo_type.addItem(t)
        self._txt_host = QtGui.QLineEdit()
        self._btn_connect = QtGui.QPushButton("Connect")
        self._btn_connect.clicked.connect(self.connect)
        
        w_connect = QtGui.QWidget()
        wc_layout = QtGui.QFormLayout()
        wc_layout.setLabelAlignment(QtCore.Qt.AlignRight)
        wc_layout.addRow("MCOR Type:", self._combo_type)
        wc_layout.addRow("Hostname/IP:", self._txt_host)
        wc_layout.addRow(self._btn_connect)
        w_connect.setLayout(wc_layout)
        size_policy = w_connect.sizePolicy()
        size_policy.setVerticalPolicy(QtGui.QSizePolicy.Fixed)
        w_connect.setSizePolicy(size_policy)
        
        self._w_sys_info = SysInfoWidget()
        self._w_xilinx_sys_monitor = XilinxSysMonitorWidget()
        self._w_v_monitor = VoltageMonitorWidget()
        
        w_basic = QtGui.QWidget()
        layout_basic = QtGui.QVBoxLayout()
        layout_basic.addWidget(w_connect)
        layout_basic.addWidget(self._w_sys_info)
        layout_basic.addWidget(self._w_xilinx_sys_monitor, stretch=2)
        layout_basic.addWidget(self._w_v_monitor, stretch=5)
        layout_basic.setMargin(0)
        w_basic.setLayout(layout_basic)
        
        self._w_channels = ChannelsControlWidget()
        
        layout = QtGui.QHBoxLayout()
        layout.addWidget(w_basic)
        layout.addWidget(self._w_channels)
        main_widget.setLayout(layout)
        
        self.setCentralWidget(main_widget)
        self.setFixedSize(1000, 650)
        self.setWindowTitle("EMCOR Controller")
        self.center()
        self.show()
    
    def center(self):
        qr = self.frameGeometry()
        cp = QtGui.QDesktopWidget().availableGeometry().center()
        qr.moveCenter(cp)
        self.move(qr.topLeft())
    
    def connect(self):
        """Try to update just the system info to test the connection."""
        with QtCore.QMutexLocker(self._exec_mutex):
            if not self._rpci:
                try:
                    self._btn_connect.setEnabled(False)
                    self._rpci = RemotePCI(self._txt_host.text())
                    # just simple tests to check if connection works
                    self._rpci.read_text(0, 1)
                    self._txt_host.setEnabled(False)
                    self._combo_type.setEnabled(False)
                    self._btn_connect.setText("Disconnect")
                except Exception as e:
                    self._rpci = None
                    msg_box = QtGui.QMessageBox()
                    msg_box.setText("Cannot connect! Check if the controller is accessible on the network and if the PCI register server is running.")
                    msg_box.setInformativeText(str(e))
                    msg_box.setIcon(QtGui.QMessageBox.Warning)
                    msg_box.exec_()
                    self._txt_host.setFocus()
                    #raise e, None, sys.exc_info()[2]
                    return
                finally:
                    self._btn_connect.setEnabled(True)
                
                mcor_type = MCOR_TYPES[str(self._combo_type.currentText())]
                self._exec_thread = QtCore.QThread()
                self._exec_obj = UpdateThreadObj(self._rpci, mcor_type, self._w_sys_info, self._w_xilinx_sys_monitor, self._w_v_monitor, self._w_channels)
                self._exec_obj.moveToThread(self._exec_thread)
                self._exec_obj.finished.connect(self._exec_thread.quit)
                self._exec_thread.started.connect(self._exec_obj.do_run)
                self._exec_thread.finished.connect(self._stop_thread_callback)
                self._exec_thread.start()
            else:
                self._exec_obj.do_stop()

    def _stop_thread_callback(self):
        with QtCore.QMutexLocker(self._exec_mutex):
            self._w_sys_info.update(None)
            self._w_xilinx_sys_monitor.update(None)
            self._w_v_monitor.update(None, None)
            self._w_channels.update(None, None)
            
            self._txt_host.setEnabled(True)
            self._combo_type.setEnabled(True)
            self._btn_connect.setText("Connect")
            
            self._rpci = None
            self._exec_thread = None
            self._exec_obj = None

def main():
    app = QtGui.QApplication(sys.argv)
    _ = MainWindow()
    return app.exec_()
    
if __name__ == '__main__':
    sys.exit(main())
