#!/usr/bin/env python

import unittest
import time
import emcor_controller as ec
import emcor_tester as et
from pyrpci import RemotePCI

IP_CONTROLLER = "134.79.218.113"
IP_TESTER = "134.79.218.58"

FULLSCALE_ADC_MON = 12
FULLSCALE_ADC_FDBCK = 12
FULLSCALE_DAC = 12.30768

MAX_DIFF_V = 0.1

SETTLE_TIME = 0.01
SLEEP_TIME_MULT = 10000

MULT_MICRO = 1000000

def configure_channel(ch_control, chnum, ramp_rate):
    """Sets fullscale values as well as ramp mode/rate and configure bit.
    If the ramp rate is 0V, immediate mode will be set.
    """
    ch_control.write_fullscale_adc_monitor(chnum, FULLSCALE_ADC_MON * MULT_MICRO)
    ch_control.write_fullscale_adc_feedback(chnum, FULLSCALE_ADC_FDBCK * MULT_MICRO)
    ch_control.write_fullscale_dac(chnum, FULLSCALE_DAC * MULT_MICRO)
    if ramp_rate == 0:
        ch_control.write_ramp_mode(chnum, True)
    else:
        ch_control.write_ramp_rate(chnum, convertVtoUA(ramp_rate))
        ch_control.write_ramp_mode(chnum, False)
    ch_control.write_configure_bit(chnum, True)

def convertVtoUA(v):
    fs_dac = int(FULLSCALE_DAC) * MULT_MICRO
    ua = ec.remap_value(v, ec.ChannelsControl.DAC_V_MIN, ec.ChannelsControl.DAC_V_MAX, -fs_dac, fs_dac)
    return ua

def convertFeedbackUAtoV(ua):
    fs_adc = FULLSCALE_ADC_FDBCK * MULT_MICRO
    v = ec.remap_value(ua, -fs_adc, fs_adc, ec.ChannelsControl.ADC_V_MIN, ec.ChannelsControl.ADC_V_MAX)
    return v

def range_float(val_min, val_max, val_step, val_fact):
    for v in range(int(val_min * val_fact), int(val_max * val_fact) + 1, int(val_step * val_fact)):
        yield float(v) / val_fact

class AIOTests(unittest.TestCase):
    def setUp(self):
        rpci = RemotePCI(IP_CONTROLLER)
        self.cc = ec.ChannelsControl(rpci)
        
        tester = et.EmcorTester(IP_TESTER)
        fault_control = et.FaultControl(tester)
        fault_control.write_all_faults(False)
        time.sleep(self.cc.TIME_CONDITION_S * SLEEP_TIME_MULT)
    
    def test_channels(self):
        """Tests that the same value can be written and read on the channels."""
        # configure channels first
        for i in range(self.cc.NUM_CHANNELS):
            configure_channel(self.cc, i, 0)
            
        for i in range(self.cc.NUM_CHANNELS):
            print "Channel: " + str(i)
            for j in range(self.cc.NUM_CHANNELS):
                self.cc.write_set_point(j, convertVtoUA(0))
            self.cc.write_set_point(i, convertVtoUA(ec.ChannelsControl.DAC_V_MAX))
            time.sleep(SETTLE_TIME)
            
            for j in range(self.cc.NUM_CHANNELS):
                test_val = 0
                if j == i:
                    test_val = ec.ChannelsControl.ADC_V_MAX
                rdbck_val = convertFeedbackUAtoV(self.cc.read_feedback(j))
                print "Channel " + str(j) + " has voltage " + str(rdbck_val)
                self.assertAlmostEqual(rdbck_val, test_val, delta=MAX_DIFF_V)
    
    def test_linearity(self):
        """Tests that generated values correspond to read values linearly."""
        val_step = 0.5
        val_fact = 10
        
        for i in range(self.cc.NUM_CHANNELS):
            # configure channel first
            configure_channel(self.cc, i, 0)
            
            print "Channel: " + str(i)
            for val in range_float(ec.ChannelsControl.DAC_V_MIN, ec.ChannelsControl.DAC_V_MAX, val_step, val_fact):
                self.cc.write_set_point(i, convertVtoUA(val))
                time.sleep(SETTLE_TIME)
                rdbck_val = convertFeedbackUAtoV(self.cc.read_feedback(i))
                print "Set V: " + str(val) + "; Read V: " + str(rdbck_val)
                # this is valid only if DAC and ADC min and max values are the same
                self.assertAlmostEqual(rdbck_val, val, delta=MAX_DIFF_V)

class AIOTimingTests(unittest.TestCase):
    TIME_SEC = 5
    MAX_DIFF_V = 0.1
    
    def setUp(self):
        rpci = RemotePCI(IP_CONTROLLER)
        self.cc = ec.ChannelsControl(rpci)
        
        tester = et.EmcorTester(IP_TESTER)
        fault_control = et.FaultControl(tester)
        fault_control.write_all_faults(False)
        time.sleep(self.cc.TIME_CONDITION_S * SLEEP_TIME_MULT)
        
    def test_linear_ramping(self):
        """Tests that linear ramping functionality works."""
        change_per_sec = float(ec.ChannelsControl.DAC_V_MAX) / self.TIME_SEC
        time_step = float(self.TIME_SEC) / 10
        time_fact = 100
        
        for i in range(self.cc.NUM_CHANNELS):
            # configure channel first for immediate mode to reset the value
            configure_channel(self.cc, i, 0)
            self.cc.write_set_point(i, convertVtoUA(0))
            time.sleep(SETTLE_TIME)
            configure_channel(self.cc, i, change_per_sec)
            
            print "Channel: " + str(i)
            t0 = time.clock()
            self.cc.write_set_point(i, convertVtoUA(ec.ChannelsControl.DAC_V_MAX))
            
            for t in range_float(time_step, self.TIME_SEC, time_step, time_fact):
                t1 = time.clock()
                t_delay = t1 - t0
                time.sleep(time_step - t_delay)
                rdbck_val = convertFeedbackUAtoV(self.cc.read_feedback(i))
                t0 = time.clock()
                print "Time: " + str(t) + "; Voltage: " + str(rdbck_val)
                val = (t / self.TIME_SEC) * ec.ChannelsControl.ADC_V_MAX
                self.assertAlmostEqual(rdbck_val, val, delta=self.MAX_DIFF_V)

class FunctionalTests(unittest.TestCase):
    def setUp(self):
        self.rpci = RemotePCI(IP_CONTROLLER)
        self.cc = ec.ChannelsControl(self.rpci)
        tester = et.EmcorTester(IP_TESTER)
        self.fault_control = et.FaultControl(tester)
    
    def test_fault_output(self):
        """Tests if channel output is set to 0 when the fault is detected on the channel."""
        for i in range(self.cc.NUM_CHANNELS):
            self.fault_control.write_all_faults(False)
            time.sleep(self.cc.TIME_CONDITION_S * SLEEP_TIME_MULT)
            
            # configure channel first
            configure_channel(self.cc, i, 0)
            
            print "Channel: " + str(i)
            self.cc.write_set_point(i, convertVtoUA(ec.ChannelsControl.DAC_V_MAX))
            self.fault_control.write_fault(True, i)
            time.sleep(self.cc.TIME_CONDITION_S * SLEEP_TIME_MULT)
            rdbck_val = convertFeedbackUAtoV(self.cc.read_feedback(i))
            self.assertAlmostEqual(rdbck_val, 0, delta=MAX_DIFF_V)
            
    def tearDown(self):
        self.fault_control.write_all_faults(False)
        time.sleep(self.cc.TIME_CONDITION_S * SLEEP_TIME_MULT)
        
        fc = ec.FaultControl(self.rpci)
        fc.reset_faults_all()
        
if __name__=='__main__':
    unittest.main(verbosity=2)
