#!/usr/bin/env python

import unittest
import emcor_controller as ec
from pyrpci import RemotePCI

IP_CONTROLLER = "134.79.218.113"

def temp_F_to_C(temp_F):
    return (temp_F - 32) * float(5) / 9

def calc_val_diff(v_in, v_req):
    v_req = float(v_req)
    return abs((v_req - v_in) / v_req)

class SystemInfoTests(unittest.TestCase):
    SYSTEM_ID = "MCOR"
    SUB_TYPE = "REV02"
    FIRMWARE_VERSION = "8.51"
    FIRMWARE_DATE = "01/15/2015"
    
    def setUp(self):
        rpci = RemotePCI(IP_CONTROLLER)
        self.sys_info = ec.SystemInfo(rpci)
    
    def test_system_id(self):
        """Tests if device has correct system id."""
        text = self.sys_info.read_system_id()
        self.assertEqual(text, self.SYSTEM_ID)
    
    def test_sub_type(self):
        """Tests if device has correct sub type."""
        text = self.sys_info.read_sub_type()
        self.assertEqual(text, self.SUB_TYPE)
    
    def test_firmware_version(self):
        """Tests if device has correct firmware version."""
        text = self.sys_info.read_firmware_version()
        self.assertEqual(text, self.FIRMWARE_VERSION)
    
    def test_firmware_date(self):
        """Tests if device has correct firmware date."""
        text = self.sys_info.read_firmware_date()
        self.assertEqual(text, self.FIRMWARE_DATE)

class XilinxSystemMonitorTests(unittest.TestCase):
    TEMP_MIN_C = temp_F_to_C(80)
    TEMP_MAX_C = temp_F_to_C(150)
    VOLT_TOLERANCE = 0.05
    
    def setUp(self):
        rpci = RemotePCI(IP_CONTROLLER)
        self.sys_mon = ec.XilinxSystemMonitor(rpci)
    
    def test_temp_cur(self):
        """Tests current temperature on Xilinx controller."""
        temp = self.sys_mon.read_temp_cur()
        print "Temperature: " + str(temp)
        self.assertGreater(temp, self.TEMP_MIN_C)
        self.assertLess(temp, self.TEMP_MAX_C)
    
    def test_temp_max(self):
        """Tests maximum temperature on Xilinx controller."""
        temp = self.sys_mon.read_temp_max()
        print "Temperature: " + str(temp)
        self.assertGreater(temp, self.TEMP_MIN_C)
        self.assertLess(temp, self.TEMP_MAX_C)
    
    def test_temp_min(self):
        """Tests minimum temperature on Xilinx controller."""
        temp = self.sys_mon.read_temp_min()
        print "Temperature: " + str(temp)
        self.assertGreater(temp, self.TEMP_MIN_C)
        self.assertLess(temp, self.TEMP_MAX_C)
    
    def test_vint_cur(self):
        """Tests INT voltage on Xilinx controller."""
        volts = self.sys_mon.read_vint_cur()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 1.0), self.VOLT_TOLERANCE)
    
    def test_vaux_cur(self):
        """Tests AUX voltage on Xilinx controller."""
        volts = self.sys_mon.read_vaux_cur()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 2.5), self.VOLT_TOLERANCE)

class VoltageMonitorTests(unittest.TestCase):
    TEMP_MIN_C = temp_F_to_C(60)
    TEMP_MAX_C = temp_F_to_C(100)
    VOLT_TOLERANCE = 0.05
    
    def setUp(self):
        rpci = RemotePCI(IP_CONTROLLER)
        self.v_mon = ec.VoltageMonitor(rpci)
    
    def test_voltage_15(self):
        """Tests +15.0V voltage on the board."""
        volts = self.v_mon.read_voltage_15()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 15.0), self.VOLT_TOLERANCE)
        
    def test_voltage_12(self):
        """Tests +12.0V voltage on the board."""
        volts = self.v_mon.read_voltage_12()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 12.0), self.VOLT_TOLERANCE)
        
    def test_voltage_5(self):
        """Tests +5.0V voltage on the board."""
        volts = self.v_mon.read_voltage_5()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 5.0), self.VOLT_TOLERANCE)
        
    def test_voltage_33(self):
        """Tests +3.3V voltage on the board."""
        volts = self.v_mon.read_voltage_33()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 3.3), self.VOLT_TOLERANCE)
        
    def test_voltage_33io(self):
        """Tests +3.3V IO voltage on the board."""
        volts = self.v_mon.read_voltage_33io()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, 3.3), self.VOLT_TOLERANCE)
        
    def test_voltage_m15(self):
        """Tests -15.0V voltage on the board."""
        volts = self.v_mon.read_voltage_m15()
        print "Voltage: " + str(volts)
        self.assertLess(calc_val_diff(volts, -15.0), self.VOLT_TOLERANCE)
    
    def test_temperature(self):
        """Tests board temperature."""
        temp = self.v_mon.read_temp()
        print "Temperature: " + str(temp)
        self.assertGreater(temp, self.TEMP_MIN_C)
        self.assertLess(temp, self.TEMP_MAX_C)

class ADCTimeoutTests(unittest.TestCase):
    def setUp(self):
        rpci = RemotePCI(IP_CONTROLLER)
        self.adc_control = ec.ADCControl(rpci)
        
    def test_adc_timeout(self):
        """Tests that ADCs are not in timeout state."""
        self.assertFalse(self.adc_control.read_adc_timeout())
        
    def test_bulk_adc_timeout(self):
        """Tests that bulk ADCs are not in timeout state."""
        self.assertFalse(self.adc_control.read_bulk_adc_timeout())
        
if __name__=='__main__':
    unittest.main(verbosity=2)
