#!/usr/bin/env python

import unittest
import time
import emcor_controller as ec
import emcor_tester as et
from pyrpci import RemotePCI

IP_CONTROLLER = "134.79.218.113"
IP_TESTER = "134.79.218.58"

SLEEP_TIME_MULT = 10000

def bits_test(func_write_one, func_write_all, func_read_reg, num_bits, sleep_time):
    stuck_lo = 0
    stuck_hi = 0xFFFF
    bad_bits = 0
    
    for bit in range(num_bits):
        func_write_all(False)
        func_write_one(True, bit)
        time.sleep(sleep_time)
        read_bits = func_read_reg()
        
        stuck_lo |= read_bits
        stuck_hi &= read_bits
        bit_mask = 1 << bit
        if bit_mask != read_bits:
            bad_bits |= bit_mask
            
    return stuck_lo, stuck_hi, bad_bits

class DITests(unittest.TestCase):
    def setUp(self):
        self.rpci = RemotePCI(IP_CONTROLLER)
        self.tester = et.EmcorTester(IP_TESTER)
        
    def check_bits(self, num_bits, stuck_lo, stuck_hi, bad_bits):
        for bit in range(num_bits):
            bit_mask = 1 << bit
            self.assertEqual(bit_mask & bad_bits, 0, "Bit " + str(bit) + " is bad")
            self.assertEqual(bit_mask & stuck_hi, 0, "Bit " + str(bit) + " is stuck high")
            self.assertGreater(bit_mask & stuck_lo, 0, "Bit " + str(bit) + " is stuck low")
    
    def test_emcor_faults(self):
        """Tests that faults are correctly detected."""
        cr = ec.FaultControl(self.rpci)
        cw = et.FaultControl(self.tester)
        
        stuck_lo, stuck_hi, bad_bits = bits_test(cw.write_fault, cw.write_all_faults, cr.read_status, cw.NUM_FAULTS, cr.TIME_CONDITION_S * SLEEP_TIME_MULT)
        self.check_bits(cw.NUM_FAULTS, stuck_lo, stuck_hi, bad_bits)
        
    def test_magnet_faults(self):
        """Tests that magnet faults are correctly detected."""
        cr = ec.MagnetFaultControl(self.rpci)
        cw = et.MagnetFaultControl(self.tester)
        
        stuck_lo, stuck_hi, bad_bits = bits_test(cw.write_mag_fault, cw.write_all_mag_faults, cr.read_status, cw.NUM_MAG_FAULTS, cr.TIME_CONDITION_S * SLEEP_TIME_MULT)
        self.check_bits(cw.NUM_MAG_FAULTS, stuck_lo, stuck_hi, bad_bits)
    
    def tearDown(self):
        fc = et.FaultControl(self.tester)
        fc.write_all_faults(False)
        mfc = et.MagnetFaultControl(self.tester)
        mfc.write_all_mag_faults(False)
        time.sleep(ec.FaultControl.TIME_CONDITION_S * SLEEP_TIME_MULT)
        
        fc = ec.FaultControl(self.rpci)
        fc.reset_faults_all()
        mfc = ec.MagnetFaultControl(self.rpci)
        mfc.reset_mag_faults_all()

class DOTests(unittest.TestCase):
    def setUp(self):
        self.rpci = RemotePCI(IP_CONTROLLER)
        self.tester = et.EmcorTester(IP_TESTER)
    
    def bit_test(self, func_write, func_read, sleep_time, invert=False):
        func_write(False)
        func_write(True)
        time.sleep(sleep_time)
        read_bit = func_read()
        
        if invert:
            self.assertEqual(read_bit, 0, "Bit should be cleared")
        else: 
            self.assertGreater(read_bit, 0, "Bit should be set")
            
        func_write(False)
        time.sleep(sleep_time)
        read_bit = func_read()
        
        if invert:
            self.assertGreater(read_bit, 0, "Bit should be set")
        else:
            self.assertEqual(read_bit, 0, "Bit should be cleared")
    
    def test_interlocks(self):
        """Tests that interlocks are correctly generated."""
        cr = et.StatusRegControl(self.tester)
        cw = ec.InterlockControl(self.rpci)
        
        stuck_lo, stuck_hi, bad_bits = bits_test(cw.write_interlock, cw.write_all_interlocks, cr.read_interlocks, cw.NUM_INTERLOCKS, cw.TIME_CONDITION_S * SLEEP_TIME_MULT)
        for bit in range(cw.NUM_INTERLOCKS):
            bit_mask = 1 << bit
            self.assertEqual(bit_mask & bad_bits, 0, "Bit " + str(bit) + " is bad")
            self.assertEqual(bit_mask & stuck_hi, 0, "Bit " + str(bit) + " is stuck high")
            self.assertGreater(bit_mask & stuck_lo, 0, "Bit " + str(bit) + " is stuck low")
        cw.write_all_interlocks(False)
    
    def test_reset_bit(self):
        """Tests that reset signal is correctly generated."""
        cr = et.StatusRegControl(self.tester)
        cw = ec.FaultControl(self.rpci)
        self.bit_test(cw.write_reset_bit, cr.read_reset_bit, cw.TIME_CONDITION_S * SLEEP_TIME_MULT)
        
    def test_inhibit_bit(self):
        """Tests that inhibit signal is correctly generated."""
        cr = et.StatusRegControl(self.tester)
        cw = ec.FaultControl(self.rpci)
        
        # set water fault to 1 (disable)
        bit = 10
        mfc = et.MagnetFaultControl(self.tester)
        address = mfc._get_address(True)
        self.tester.write(address, [(1 << bit) & 0xFFFF])
        self.bit_test(cw.write_inhibit_bit, cr.read_inhibit_bit, cw.TIME_CONDITION_S * SLEEP_TIME_MULT, invert=True)

if __name__=='__main__':
    unittest.main(verbosity=2)
